/*
Терморегулятор для дистилятора (з оптопарою РС814 замість РС817, без діодного моста)
*/

#include <EEPROM.h>
#include <OneWire.h>
#include <LiquidCrystal.h>

#define pin_A 2  // pin 2 - поворот енкодера (переривання, зріз імпульсу)
#define pin_B 8 // pin 8 - напрямок обертання енкодера
#define pin_sw 9 // pin 9 - кнопка енкодера

#define mode_sw 10 // pin 10 - перемикач режимів
#define zero_c 3  // pin 3 - детектор нуля (переривання, зріз імпульсу)
#define pulse_out 11 // pin 11 - вихід імпульсу відпирання симістора

#define pulseHigh(pin) {digitalWrite(pin, HIGH); delayMicroseconds(5);digitalWrite(pin, LOW); }
#define change_step 0.1 //крок змінення завдання по температурі

#ifndef TCKCSR
#define TCKCSR (*((volatile unsigned char *)0xEC))//for non-LGT boards compatibility
#endif

float Kp; //коефіцієнт передачі ПІД регулятора
float Ti; //час ізодрому інтегральної ланки ПІД регулятора
float Td; //час діференціювання ПІД регулятора

byte qt_addr[8]; //адреса датчика температури
byte mode;//режим регулятора (0-ручний, 1-авто)
byte counter = 0;//лічильник параметрів для інженерного меню
byte edit = 0;//режим редагування в інженерному меню
byte set = 0;//1-режим інженерного меню
byte first_way = 1;
unsigned int phase_us=10000; //кут керування симістором (затримка відпирання в мікросекундах)
float OP0;//початкова точка регулятора
float OP; //потужність нагрівача у відсотках
float Temp_SP;    // завдання для регулятора температури
volatile byte flag_interrupt_turn = 0;
unsigned long time_s, time_r;
float Temperature;    // поточна виміряна температура
float Integral = 0;
float Diff = 0;
float Err = 0;
const char signature[] = "D.Ledovyi  2020";
float *Value_addr[4] = {&Kp, &Ti, &Td, &OP0};

OneWire  ds(12);  // датчик температури підключено до pin 12 (4.7K резистор на плюс живлення)
LiquidCrystal lcd(1, 0, 4, 5, 6, 7);
 
void setup()
{ 
  pinMode(pin_A, INPUT);// для читання з виходу А енкодера
  digitalWrite(pin_A, HIGH);// вмикаємо підтягуючий резистор
  pinMode(pin_B, INPUT);// для читання з виходу В енкодера
  digitalWrite(pin_B, HIGH);// вмикаємо підтягуючий резистор
  pinMode(pin_sw, INPUT);// для читання з кнопки енкодера
  digitalWrite(pin_sw, HIGH);// вмикаємо підтягуючий резистор
  pinMode(mode_sw, INPUT);// для читання з перемикача режимів
  digitalWrite(mode_sw, HIGH);// вмикаємо підтягуючий резистор
  pinMode(zero_c, INPUT);// для читання з детектора нуля
  digitalWrite(zero_c, HIGH);// вмикаємо підтягуючий резистор
  pinMode(pulse_out, OUTPUT);// для видачі імпульса відпирання симістора

  /*EEPROM_float_write(0, 29.0);//перший запуск на чистий контроллер
    EEPROM_float_write(4, 20.0);
    EEPROM_float_write(8, 10000.0);
    EEPROM_float_write(12, 0.0);
    EEPROM_float_write(16, 10.0);*/
   
  
  Kp = EEPROM_float_read(4);
  Ti = EEPROM_float_read(8);
  Td = EEPROM_float_read(12);
  OP0 = EEPROM_float_read(16);
  
  lcd.begin(16, 2);
  lcd.setCursor(0, 0);
  lcd.print("TempReg ver 4.6");
  lcd.setCursor(0, 1);
  lcd.print(signature);
  delay(2000);

  lcd.clear();

  if(!digitalRead(pin_sw))//якщо натиснуто кнопку енкодера, режим налаштування
    {
      set = 1;
      lcd.setCursor(0, 0);
      lcd.print("SETUP...");
      lcd.setCursor(0, 1);
      lcd.print("Kp");
      while(!digitalRead(pin_sw));//чекаємо на відпускання кнопки
      delay(200);
      time_s = millis();
      attachInterrupt(0, enc_turn, FALLING); //вихід А енкодера підключено до пін 2 (int 0). Чекаємо зріз імпульса
      return;
    }

  OP = OP0;
 
  Temp_SP = EEPROM_float_read(0);
  
  timer_ini();
  
  while(!qt_setup())
  {
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("Sensor error.");
    ds.reset_search();
    delay(1000);
    if (!digitalRead(mode_sw))
       break;
  }
 
  //запуск першого вимірювання датчиком
  ds.reset();
  ds.select(qt_addr);
  ds.write(0x44, 1);        // start conversion, with parasite power on at the end
  
  time_s = millis();
  time_r = time_s;
   
  attachInterrupt(0, enc_turn, FALLING); //вихід А енкодера підключено до пін 2 (int 0). Чекаємо зріз імпульса
  attachInterrupt(1, zero_detect, FALLING); //вихід детектора нуля підключено до пін 3 (int 1). Чекаємо зріз імпульса
} 
 
void loop()
{
 if(set)
   {
     setting();
     return;
   }
   
 if(!digitalRead(pin_sw))//якщо натиснуто кнопку енкодера
   {
     EEPROM_float_write(0, (float)(int((Temp_SP)*10))/10);//зберігаємо завдання по температурі в енергонезалежну пам"ять
     lcd.setCursor(0, 1);
     lcd.print("***SAVED***");
     delay(3000);
     while(!digitalRead(pin_sw));//чекаємо на відпускання кнопки
     lcd_print_SP();
   }
   
 if(flag_interrupt_turn)
   check_flag_turn();

 mode = digitalRead(mode_sw);//режим

 if((millis() - time_s) >= 913)
 {
   //зчитуємо показники температури
   float New_Temperature = qt_read();
   if((abs(New_Temperature - Temperature) < 20) || first_way)//Якщо збій датчика, залишаємо старе значення
      Temperature = New_Temperature;
   first_way = 0;

   //запуск нового вимірювання датчиком
   ds.reset();
   ds.select(qt_addr);
   ds.write(0x44, 1);        // start conversion, with parasite power on at the end
   time_s = millis();
   
   //виводимо температуру на РКД     
   lcd.setCursor(0, 0);
   lcd.print("                ");
   lcd.setCursor(0, 0);
   lcd.print("PV=");
   lcd.print(Temperature, 2);
   lcd.write(223);
   lcd.print("C  ");
   lcd.print("OP");
   if(!mode)//якщо ручний режим
     lcd.print("*");
 }

 if((millis() - time_r) >= 1000)//раз на секунду розраховуємо фазову затримку та вихідну потужність
 {
  time_r = millis();
  if(mode)//якщо атоматичний режим 
  {
     if(Temperature >= 0)//якщо підозра на обрив датчика, розрахунок не робимо
     {
       Diff = Td * (Temp_SP - Temperature-Err);
       Err = Temp_SP - Temperature;
       if(Diff > 1000.0)
          Diff = 1000.0;
       if(Diff < -1000.0)
          Diff = -1000.0;
     
       if(Ti > 0)
       {
        Integral += Err/Ti;
       
        if(Integral > (100-OP0)/Kp)
           Integral = (100-OP0)/Kp;
        if(Integral < -OP0/Kp)
           Integral = -OP0/Kp;
       }
       else Integral = 0;
        
     
       OP = OP0 + Kp * (Err + Integral + Diff);
       if(OP > 100.0)
          OP = 100.0;
     }

  }
  else//якщо ручний режим
  {
     OP0 = OP;
     Temp_SP = Temperature;
     Integral = 0.0;
     Err = 0.0;
  }
 if(OP < 0.0)
     OP = 0.0;
 lcd_print_SP();
 phase_us = int((100.0 - OP)*100.0); //фазова затримка міняється від 0 (потужність 100%) до 10000 мкс (потужність 0)
 phase_us += 270;//імпульс переривання від детектора нуля надходить трохи раніше реального нуля, тому +поправка
 }
}

void enc_turn()//обробник переривань від обертання енкодера
{
  noInterrupts();//заборона переривань для запобігання брязкіту
  flag_interrupt_turn = 1;
  return;
}

void zero_detect()//обробник переривань від детектора нуля
{
  if(phase_us > 9500)//Якщо потужність менше 5%, імпульс не видається
    return;
  if(mode && (Temperature < 0))//Якщо в автоматичному режимі обірвано датчик, імпульс не видається
    return;
  TIFR1=TIFR1;
  OCR1A = int(phase_us*2); //записуємо в регістр число для порівняння, один такт таймера - 0.5 мікросекунди
  TIMSK1 |= (1<<OCIE1A); //встановлюємо біт дозволу переривання 1ого таймера по співпадінню з OCR1A
  TCCR1B |= (1<<CS11);; //пуск таймера з подільником частоти на 8!
  return;
}

void check_flag_turn()//перевірка прапорця обертання енкодера
{
  if(mode)//якщо атоматичний режим
  {
    delay(2);//затримка перед обробкою переривання для запобігання брязкіту
    if(digitalRead(pin_A))//якщо вихід А високий, означає, що пререривання хибне
      {
        flag_interrupt_turn = 0;
        interrupts(); //дозвіл переривань
        return;
      }

    if(digitalRead(pin_B))
      {
        // вихід В в стані 1, означає, що обертання за годинниковою стрілкою
        // збільшуємо завдання на change_step, але не більше, ніж до 120 градусів
        if(Temp_SP + change_step <= 120) 
          Temp_SP += change_step;
        else Temp_SP = 120;              
      }   
    else 
      {
        // вихід В в 0 стані, означає, що обертання проти годинникової стрілки     
        // зменшуємо завдання на change_step, але не менше, ніж до 0 градусів
        if(Temp_SP - change_step >= 0)
          Temp_SP -= change_step; 
        else Temp_SP = 0;               
      }
  }
  else//якщо ручний режим
  {
    delay(2);//затримка перед обробкою переривання для запобігання брязкіту
    if(digitalRead(pin_A))//якщо вихід А високий, означає, що пререривання хибне
      {
        flag_interrupt_turn = 0;
        interrupts(); //дозвіл переривань
        return;
      }

    if(digitalRead(pin_B))
      {
        // вихід В в стані 1, означає, що обертання за годинниковою стрілкою
        // збільшуємо вихідну потужність на 1%, але не більше, ніж до 100%
        if(OP + 1 <= 100) 
          OP += 1;
        else OP = 100;              
      }   
    else 
      {
        // вихід В в 0 стані, означає, що обертання проти годинникової стрілки     
        // зменшуємо вихідну потужність на 1%, але не менше, ніж до 0%
        if(OP - 1 >= 0)
          OP -= 1; 
        else OP = 0;               
      }
  }
    lcd_print_SP();
    flag_interrupt_turn = 0;
    interrupts(); //дозвіл переривань
    return;
}

void lcd_print_SP()//виведення завдання на РКІ
{
  lcd.setCursor(0, 1);
  lcd.print("                ");
  lcd.setCursor(0, 1);
  lcd.print("SP=");
  lcd.print(Temp_SP, 2);
  lcd.write(223);
  lcd.print("C");
  lcd.setCursor(12, 1);
  lcd.print((int)OP);
  lcd.setCursor(15, 1);
  lcd.print("%");

  return;
}

byte qt_setup(void)
{
  if (!ds.search(qt_addr)) 
    return(0);
  
  if (OneWire::crc8(qt_addr, 7) != qt_addr[7])
    return(0);
    
  return(1);
}

float qt_read(void)
{
  byte i;
  byte present = 0;
  byte type_s;
  byte data[12];
  
  present = ds.reset();
  ds.select(qt_addr);    
  ds.write(0xBE);         // Read Scratchpad
 
  for ( i = 0; i < 9; i++)          // we need 9 bytes
    data[i] = ds.read();
      
  // Convert the data to actual temperature
  // because the result is a 16 bit signed integer, it should
  // be stored to an "int16_t" type, which is always 16 bits
  // even when compiled on a 32 bit processor.
  int16_t raw = (data[1] << 8) | data[0];
  if (type_s) {
    raw = raw << 3; // 9 bit resolution default
    if (data[7] == 0x10) {
      // "count remain" gives full 12 bit resolution
      raw = (raw & 0xFFF0) + 12 - data[6];
    }
  } else {
    byte cfg = (data[4] & 0x60);
    // at lower res, the low bits are undefined, so let's zero them
    if (cfg == 0x00) raw = raw & ~7;  // 9 bit resolution, 93.75 ms
    else if (cfg == 0x20) raw = raw & ~3; // 10 bit res, 187.5 ms
    else if (cfg == 0x40) raw = raw & ~1; // 11 bit res, 375 ms
    //// default is 12 bit resolution, 750 ms conversion time
  }
  return((float)raw / 16.0);
}

void EEPROM_float_write(int addr, float val) // запис в EEPROM
{ 
  byte *x = (byte *)&val;
  for(byte i = 0; i < 4; i++) EEPROM.write(i+addr, x[i]);
}

float EEPROM_float_read(int addr) // читання з EEPROM
{   
  byte x[4];
  for(byte i = 0; i < 4; i++) x[i] = EEPROM.read(i+addr);
  float *y = (float *)&x;
  return y[0];
}

void timer_ini(void)
{
  TCKCSR=0; TIMSK1=0; //обнулити зайве
  TCCR1A=0; TCCR1B=0; //регісти управління, обнулити зайве
  TCCR1B |= (1<<WGM12); // встановлюємо режим СТС (скидання за співпадінням) 
  OCR1A=0; OCR1B=65535u;// регістри порівняння
  return;
}

ISR (TIMER1_COMPA_vect)//прийшло переривання від таймера
{
  TIMSK1 = 0;//заборона переривань від таймерів
  TCCR1B&=252;//стоп таймер
  pulseHigh(pulse_out);//видача відпираючого імпульсу на симістор
  return;
} 

void setting(void)
{
   if(flag_interrupt_turn)
   {
     if(edit)
       check_flag_turn_edit();
       
     else
       check_flag_turn_counter();
       
     flag_interrupt_turn = 0;
     interrupts(); //дозвіл переривань
   }
   
   if(!digitalRead(pin_sw))//якщо натиснуто кнопку енкодера
   {
    if(edit)
      {
        EEPROM_float_write(counter*4 + 4, *Value_addr[counter]);
        lcd.setCursor(0, 1);
        edit = 0;
        lcd.print("***SAVED***");
        delay(3000);
        while(!digitalRead(pin_sw));//чекаємо на відпускання кнопки
      }
     else 
      {
        edit = 1;
        delay(300);
        while(!digitalRead(pin_sw));//чекаємо на відпускання кнопки
      }
   }

   if((millis() - time_s) >= 300)
   {
     lcd.setCursor(0, 1);
     lcd.print("           ");
     lcd.setCursor(0, 1);
   
   
     if(edit)
       {
         switch(counter)
           {
             case 0:
               lcd.print(Kp, 1);
               break;
 
             case 1:
               lcd.print(int(Ti));
               break;

             case 2:
               lcd.print(Td, 1);
               break;

             case 3:
               lcd.print(int(OP0));
           }
       }
       
    else
       {
         switch(counter)
           {
             case 0:
               lcd.print("Kp=");
               lcd.print(Kp, 1);
               break;
 
             case 1:
               lcd.print("Ti=");
               lcd.print(int(Ti));
               break;

             case 2:
               lcd.print("Td=");
               lcd.print(Td, 1);
               break;

             case 3:
               lcd.print("OP0=");
               lcd.print(int(OP0));
           }
       }
   time_s = millis();
   }
   return;
}

void check_flag_turn_counter(void)
{
  delay(2);//затримка перед обробкою переривання для запобігання брязкіту
  
  if(digitalRead(pin_A))//якщо вихід А високий, означає, що пререривання хибне
    return;

  if(digitalRead(pin_B))
    {
      // вихід В в стані 1, означає, що обертання за годинниковою стрілкою
      // збільшуємо лічильник на 1 по колу 0-3-0
      if(counter + 1 <= 3) 
        counter += 1;
      else counter = 0;              
    }   
  else 
    {
      // вихід В в 0 стані, означає, що обертання проти годинникової стрілки     
      // зменшуємо лічильник на 1 по колу 3-0-3
      if(counter - 1 >= 0)
        counter -= 1; 
      else counter = 3;               
    }
  return; 
}

void check_flag_turn_edit(void)
{
  const float ch_steps[4] = {0.1, 100.0, 0.1, 1.0};
  const float maxes[4] = {100.0, 100000.0, 100.0, 100.0};
  
  delay(2);//затримка перед обробкою переривання для запобігання брязкіту
  
  if(digitalRead(pin_A))//якщо вихід А високий, означає, що пререривання хибне
    return;

  if(digitalRead(pin_B))
    {
      // вихід В в стані 1, означає, що обертання за годинниковою стрілкою
      // збільшуємо параметр на його ch_step, але не більше max
      if(*Value_addr[counter] + ch_steps[counter] <= maxes[counter]) 
        *Value_addr[counter] += ch_steps[counter];
      else *Value_addr[counter] = maxes[counter];              
    }   
  else 
    {
      // вихід В в 0 стані, означає, що обертання проти годинникової стрілки     
      // зменшуємо параметр на його ch_step, але не менше 0
      if(*Value_addr[counter] - ch_steps[counter] >= 0) 
        *Value_addr[counter] -= ch_steps[counter];
      else *Value_addr[counter] = 0;            
    }
  return; 
}
